/*******************************************************************
  config.c  --  Copyright 1999 - 2003,
                   Alec Kosky (aleckosky@hotmail.com)

Alright, here's the deal (i.e. the AK License):

You are free to use this program for commercial as well as 
   non-commercial endeavors.
You are free to modify this program to suit your needs.
You are free to redistribute unmodified copies of this program.

Basically, you can do what you like with it. However, you are not
permitted to redistribute modified copies of this program. If you
would like to see changes made to this program, please submit
them to me (aleckosky@hotmail.com) and I will incorporate them into
future versions if I feel that they are beneficial.

********************************************************************/
#include "connlogd.h"

/*
 * Externally-defined variables and functions.
 */
extern int log_level, log_iponly, log_debug;

extern int host_clear(void);
extern int host_insert(unsigned long);
extern int d_port_clear(void);
extern int d_port_insert(unsigned int);
extern int s_port_insert(unsigned long int host, unsigned int port);
extern int log_conn(int, const char *, ...);


struct name_value syslog_facilities[] =
{
   { "stderr",       0 },
   { "daemon",       LOG_DAEMON },
   { "auth",         LOG_AUTH },
   { "mail",         LOG_MAIL },
   { "user",         LOG_USER },
   { "local0",       LOG_LOCAL0 },
   { "local1",       LOG_LOCAL1 },
   { "local2",       LOG_LOCAL2 },
   { "local3",       LOG_LOCAL3 },
   { "local4",       LOG_LOCAL4 },
   { "local5",       LOG_LOCAL5 },
   { "local6",       LOG_LOCAL6 },
   { "local7",       LOG_LOCAL7 },
   { CHAR_NULL,      -1 }
};


struct name_value syslog_levels[] = 
{
   { "emerg",     LOG_EMERG },
   { "alert",     LOG_ALERT },
   { "crit",      LOG_CRIT },
   { "err",       LOG_ERR },
   { "warning",   LOG_WARNING },
   { "notice",    LOG_NOTICE },
   { "info",      LOG_INFO },
   { "debug",     LOG_DEBUG },
   { CHAR_NULL,   -1 }
};


/*
 * Allow us to read entire lines from a file descriptor.
 */
ssize_t readn(int fd, char *buf, size_t buflen)
{
   int       i;
   char     *ptr, tmp[3];
   ssize_t   nread = 0, cnt = 0;

   ptr = buf;
   while (nread < buflen - 1)
   {
      i = 0;
      do_read:
      if ((cnt = read(fd, tmp, 1)) <= 0)
      {
         if (errno != EINTR && errno != EAGAIN)
            return (-1);
         else
         {
            if (i++ >= 20)
               return (-1);
            else
               goto do_read;
         }
      }
      else if (tmp[0] == '\n')
         break;
      else
         *ptr++ = tmp[0];
      nread += cnt;
   }

   /*
    * Ensure that we've got a terminating null character.
    */
   buf[buflen] = '\0';
   return (nread);
}

/*
 * This function will pull out the next 'word' (separated by whitespace)
 * from the given string, starting after the specified index.
 */
int find_word(const char *input, int index, char *buf, int buflen)
{
   int   i, b, found;

   /*
    * Make sure that we aren't already at the end of the string.
    */
   if (index < 0 || input[index] == '\0')
      return (-1);

   /*
    * Now find the next word in the given string (skip the
    * extra whitespace stuff).
    */
   b = 0;
   i = index;
   found = 0;
   while (input[i] != '\0')
   {
      if (input[i] == ' '  || input[i] == '\t' || input[i] == '\n')
      {
         if (found == 0)
            i++;         /* move on to the next character */
         else
            break;       /* this is the space after the word */
      }
      else
      {
         found = 1;
         if (b < buflen - 2)
            buf[b++] = input[i];
         i++;            /* move on to the next character */
      }
   }

   /*
    * Add the terminating null-character.
    */
   buf[b] = '\0';

   /*
    * Return the index (useful for future calls to find the next string.
    */
   return (i);
}


/*
 * This is the function that will parse the configuration
 * file into meaningul variables.
 */
int parse_cfg(int fd)
{
   unsigned long host;
   unsigned int  portval;
   int           comment, line, val, idx, i, j;
   char          buf[256], tmp[128], *ctmp;

   if (fd <= 0)
   {
      log_conn(0, "parse_cfg: invalid file descriptor");
      return (-1);
   }

   /*
    * Read each line, then see if we should parse it.
    * Clean out all previously allocated memory before starting.
    */
   val  = 0;
   host = 0;
   line = 0;
   log_level = 0;
   log_debug = 0;
   log_iponly = 1;
   host_clear();
   d_port_clear();
   memset(buf, 0, sizeof(buf));
   while (readn(fd, buf, sizeof(buf) - 2) >= 0)
   {
      line++;
      comment = 0;
      if (buf[0] != '#' && buf[0] != '\0')
      {
         if ((ctmp = strchr(buf, '=')) == NULL)
         {
            log_conn(0, "ERROR: On line %d: missing '='", line);
            return (-1);
         }

         idx = ctmp - (char *)&buf + 1;
         if (strncasecmp("ign_host", buf, 8) == 0)
         {
            /*
             * OK, we have a host that we want to ignore. Let's read it
             * in, then read in the list of source ports to ignore.
             */
            i = 0;
            memset(tmp, 0, sizeof(tmp));
            while ((idx = find_word(buf, idx, tmp, sizeof(tmp))) >= 0)
            {
               if (strchr(tmp, '#') != NULL)
                  comment = 1;
               else if (comment == 0)
               {
                  /*
                   * We don't have any comments in this string.
                   */
                  if (i++ == 0)
                  {
                     /*
                      * Check to see if this is the * character - meaning
                      * all hosts.
                      */
                     if (tmp[0] == '*')
                        host = 0;
                     else
                        host = inet_addr(tmp);

                     /*
                      * Alright, we've now got the host - add it to the list
                      */
                     if ((val = host_insert(host)) != 0)
                     {
                        log_conn(0,
             "WARNING: Line %d: cannot add host \"%s\" to list of hosts (%s)",
                                 line, tmp, val == 1 ? "already in list" :
                                 "memory allocation error");
                        return (-1);
                     }
                  }
                  else if (tmp[0] != '\0')
                  {
                     /*
                      * Now we have a port to ignore for this host.
                      */
                     if ((portval = (unsigned int)strtol(tmp, (char **)NULL,
                                                         10)) <= 0)
                     {
                        log_conn(0,
                    "ERROR: On line %d: error converting %s to a port number",
                               line, tmp);
                        return (-1);
                     }

                     if ((val = s_port_insert(host, portval)) != 0)
                     {
                        if (val == -2)
                           log_conn(0,
              "ERROR: On line %d: cannot add source port %s for unknown host",
                                    line, tmp);
                        else
                           log_conn(0,
               "WARNING: Line %d: cannot add %s to list of source ports (%s)",
                                    line, tmp, val == 1 ? "already in list" :
                                    "memory allocation error");
                        return (-1);
                     }
                  }
               }
               memset(tmp, 0, sizeof(tmp));
            }
         }
         else if (strncasecmp("ign_dport", buf, 9) == 0)
         {
            /*
             * Here's a destination port that we don't care about.
             */
            memset(tmp, 0, sizeof(tmp));
            while ((idx = find_word(buf, idx, tmp, sizeof(tmp))) >= 0)
            {
               if (strchr(tmp, '#') != NULL)
                  comment = 1;
               else if (comment == 0)
               {
                  /*
                   * Ensure that we have a valid port (as an
                   * unsigned integer (and stop reading if we encounter
                   * a comment).
                   */
                  if (strchr(tmp, '#') != NULL)
                     comment = 1;
                  else if (comment == 0)
                  {
                     if ((portval = (unsigned int)strtol(tmp, (char **)NULL,
                                                         10)) > 0)
                     {
                        if (d_port_insert((unsigned int)portval) < 0)
                        {
                           log_conn(0,
          "WARNING: Line %d: cannot add %s to list of destination ports (%s)",
                                    line, tmp, val == 1 ? "already in list" :
                                    "memory allocation error");
                           return (-1);
                        }
                     }
                  }
               }
               memset(tmp, 0, sizeof(tmp));
            }
         }
         else if (strncasecmp("log_hname", buf, 9) == 0)
         {
            /*
             * Only log the IP address (don't do the DNS lookup).
             */
            memset(tmp, 0, sizeof(tmp));
            if (find_word(buf, idx, tmp, sizeof(tmp)) < 0)
            {
               log_conn(0,
                   "ERROR: On line %d: could not find argument for log_hname",
                        line);
               return (-1);
            }

            if (strncasecmp("on", tmp, 2) == 0  ||
                strncasecmp("yes", tmp, 3) == 0 ||
                strncasecmp("true", tmp, 4) == 0)
               log_iponly = 0;
            else
               log_iponly = 1;
         }
         else if (strncasecmp("log_style", buf, 9) == 0)
         {
            /*
             * We are parsing the log configuration.
             */
            i = 0;
            memset(tmp, 0, sizeof(tmp));
            while ((idx = find_word(buf, idx, tmp, sizeof(tmp))) >= 0 &&
                   i < 2)
            {
               if (strchr(tmp, '#') != NULL)
                  comment = 1;
               else if (comment == 0)
               {
                  if (i == 0)
                  {
                     /*
                      * Check to see if we're doing syslog logging.
                      */
                     for (j = 0; syslog_facilities[j].name != CHAR_NULL; j++)
                     {
                        if (strncasecmp(syslog_facilities[j].name, tmp,
                                     strlen(syslog_facilities[j].name)) == 0)
                        {
                           log_level = syslog_facilities[j].value;
                           break;
                        }
                     }

                     if (syslog_facilities[j].name == CHAR_NULL)
                     {
                        /*
                         * We are doing standard error logging (not syslog).
                         */
                        log_level = 0;
                        i = 2;
                     }
                  }
                  else if (i == 1)
                  {
                     /*
                      * Syslog logging - let's see if we have a log level.
                      */
                     for (j = 0; syslog_levels[j].name != CHAR_NULL; j++)
                     {
                        if (strncasecmp(syslog_levels[j].name, tmp,
                                        strlen(syslog_levels[j].name)) == 0)
                        {
                           log_level |= syslog_levels[j].value;
                           break;
                        }
                     }
                  }
                  i++;
               }
               memset(tmp, 0, sizeof(tmp));
            }
         }
         else if (strncasecmp("log_debug", buf, 9) == 0)
         {
            /*
             * Setup extra verbose logging at run-time
             * (no longer has to be set at compile-time).
             */
            memset(tmp, 0, sizeof(tmp));
            if (find_word(buf, idx, tmp, sizeof(tmp)) < 0)
               return (-1);

            if (strncasecmp("on", tmp, 2) == 0  || 
                strncasecmp("yes", tmp, 3) == 0 ||
                strncasecmp("true", tmp, 4) == 0)
               log_debug = 1;
            else if (strncasecmp("verbose", tmp, 7) == 0)
               log_debug = 2;
         }
         else
         {
            log_conn(0, "WARNING: Line %d: ignoring unknown command: '%s'",
                     buf);
         }
      }
      memset(buf, 0, sizeof(buf));
   }

   return (0);
}


/*
 * This is the basic wrapper function.
 */
int read_cfg(char *file)
{
   int    fd, rval;
   
   if (file == NULL)
   {
      log_conn(0, "ERRPR: read_cfg: cannot read file (NULL)");
      return (-1);
   }

   if ((fd = open(file, O_RDONLY)) < 0)
   {
      log_conn(0, "ERROR: read_cfg: cannot open %s (%s)", file,
               strerror(errno));
      return (-1);
   }

   rval = parse_cfg(fd);
   close (fd);

   if (rval != 0)
      return (-1);
   else
      return (0);
}


