/*******************************************************************
 connlogd.c  --  Copyright 1999 - 2003, 
                    Alec Kosky (aleckosky@hotmail.com)

Alright, here's the deal (i.e. the AK License):

You are free to use this program for commercial as well as 
   non-commercial endeavors.
You are free to modify this program to suit your needs.
You are free to redistribute unmodified copies of this program.

Basically, you can do what you like with it. However, you are not
permitted to redistribute modified copies of this program. If you
would like to see changes made to this program, please submit
them to me (aleckosky@hotmail.com) and I will incorporate them into
future versions if I feel that they are beneficial.

********************************************************************/
#include "connlogd.h"

/*
 * We need the pcap variable to be global in order to close
 * the packet capture device in the signal handler.
 */
int                log_level, log_iponly, log_debug;
char               cfgfile[256];
pcap_t            *pcap;
struct udpconn_t  *udplist;

extern  int     clean_udplist(void);
extern  int     host_clear(void);
extern  int     d_port_clear(void);
extern  int     log_conn(int, const char *, ...);
extern  void    log_exit(int, const char *, ...);
extern  int     read_cfg(char *);
extern  int     read_packets(pcap_t *);
extern  pcap_t *pcap_open(char *, int, int);

/*
 * A simple signal handler to cleanup and terminate.
 */
void sigerr_handler(int signum)
{
   struct udpconn_t  *ptr, *ptr1;
   sigset_t           newmask;

   /*
    * Block *ALL* signals to prevent any signal-delivery problems.
    */
   sigfillset(&newmask);
   if (sigprocmask(SIG_SETMASK, &newmask, NULL) != 0)
      return;

   /*
    * Close the packet capture device.
    */   
   pcap_close(pcap);
   ptr = udplist;
   while (ptr != NULL)
   {
      ptr1 = ptr;
      ptr = ptr->next;
      free(ptr1);
   }
   log_conn(log_level,
            "NOTICE: Connlogd exiting upon receipt of signal #%d", signum);
   if (signum == SIGTERM)
      exit (0);
   else
      exit (1);
}

/*
 * This is the signal handler to allow us to cleanup
 * the UDP connection data structures.
 */
void sigalrm_handler(int signum)
{
   sigset_t    newmask, oldmask;

   /*
    * Block *ALL* signals to prevent any signal-delivery problems.
    */
   sigfillset(&newmask);
   if (sigprocmask(SIG_SETMASK, &newmask, &oldmask) != 0)
      return;

   clean_udplist();
   alarm(UDP_LIFE);		/* re-clean the list in the future */

   /*
    * Restore the original signal mask.
    */
   sigprocmask(SIG_SETMASK, &newmask, &oldmask);
}

/*
 * Handle the HANGUP signal (SIGHUP) - tells us
 * to re-read the configuration file. First the memory
 * that had been allocated is cleared, then the config
 * file is re-read. If it cannot successfully re-read
 * the config file, then it will give an error message
 * to auth.warn (via syslog), and clear everything out
 * of memory (thus nothing will be ignored because
 * someone screwed up the config file).
 */
void sighup_handler(int signum)
{
   int           i;
   struct stat   filestat;
   sigset_t      newmask, oldmask;

   /*
    * Block *ALL* signals to prevent any signal-delivery problems.
    */
   sigfillset(&newmask);
   if (sigprocmask(SIG_SETMASK, &newmask, &oldmask) != 0)
      return;

   /*
    * Re-read the configuration file.
    */
   if (cfgfile[0] != 0)
   {
      i = 0;
      log_conn(log_level, "CONFIG: Reloading configuration from %s", cfgfile);

      my_dostat:
      if (lstat(cfgfile, (struct stat *)&filestat) == -1)
      {
         if (errno == EINTR || errno == EAGAIN)
         {
            if (i++ < 20)
               goto my_dostat;
            else
               log_conn((LOG_AUTH | LOG_WARNING), 
                        "WARNING: Could not stat %s after %d tries.",
                        cfgfile, i);
         }
      }
      else
      {
         if (read_cfg(cfgfile) != 0)
         {
            log_conn(log_level,
                     "WARNING: Error while reading configuration file: %s\n",
                     cfgfile);
            host_clear();
            d_port_clear();
         }
      }
   }

   /*
    * Now we can unblock signal-delivery.
    */
   sigprocmask(SIG_SETMASK, &oldmask, NULL);
}

/*
 * A simple function to set up signal handlers and prevent core dumps.
 */
void setup_handlers(void)
{
   struct sigaction   term, alrm, hup;

   /*
    * If we have setrlimit, disallow core dumps (add a second
    * level of core dump protection, since we're going to
    * be handling most of the possible core-dump signals).
    */
#if defined(HAVE_SETRLIMIT) && defined(RLIMIT_CORE)
   struct rlimit  rl;

   getrlimit(RLIMIT_CORE, (struct rlimit *)&rl);
   rl.rlim_cur = 0;
   if (setrlimit(RLIMIT_CORE, (struct rlimit *)&rl) != 0)
      log_exit(0, "ERROR: Could not limit core dumps.");
#endif

   /*
    * Handle the various error/termination signals - we
    * need to close the packet capture device and exit
    * gracefully.
    */
   term.sa_handler = sigerr_handler;
   term.sa_flags   = 0;
   /*
    * Ignore the signal to prevent recursive signal problems.
    */
   sigemptyset(&term.sa_mask);
   sigaddset(&term.sa_mask, SIGTERM);
   sigaddset(&term.sa_mask, SIGINT);
   sigaddset(&term.sa_mask, SIGQUIT);
   sigaddset(&term.sa_mask, SIGILL);
   sigaddset(&term.sa_mask, SIGABRT);
   sigaddset(&term.sa_mask, SIGFPE);
   sigaddset(&term.sa_mask, SIGBUS);
   sigaddset(&term.sa_mask, SIGSEGV);
   sigaction(SIGTERM, (struct sigaction *)&term, NULL);
   sigaction(SIGINT,  (struct sigaction *)&term, NULL);
   sigaction(SIGQUIT, (struct sigaction *)&term, NULL);
   sigaction(SIGILL,  (struct sigaction *)&term, NULL);
   sigaction(SIGABRT, (struct sigaction *)&term, NULL);
   sigaction(SIGFPE,  (struct sigaction *)&term, NULL);
   sigaction(SIGBUS,  (struct sigaction *)&term, NULL);
   sigaction(SIGSEGV, (struct sigaction *)&term, NULL);

   /*
    * Handle SIGALRM signals
    */
   alrm.sa_handler = sigalrm_handler;
   alrm.sa_flags   = SA_RESTART;
   /*
    * Ignore the signal to prevent recursive signal problems.
    */
   sigemptyset(&alrm.sa_mask);
   sigaddset(&alrm.sa_mask, SIGALRM);
   sigaction(SIGALRM, (struct sigaction *)&alrm, NULL);

   /*
    * Handle SIGHUP signals
    */
   hup.sa_handler = sighup_handler;
   hup.sa_flags   = SA_RESTART;
   /*
    * Ignore the signal to prevent recursive signal problems.
    */
   sigemptyset(&hup.sa_mask);
   sigaddset(&hup.sa_mask, SIGHUP);
   sigaction(SIGHUP, (struct sigaction *)&hup, NULL);
}

int background(int do_fork)
{
   int  pid;

   /*
    * Check to see if we our parent process is the
    * init process (we might not have to do the rest).
    */
   nice(15);
   if (getppid() != 1)
   {
      /*
       * Let's daemonize ourself (make us into
       * a background process) so that we can
       * continue without being a pest.
       */
      if (do_fork != 0)
      {
         if ((pid = fork()) == -1)
         {
            log_conn(0, "ERROR: Could not successfully fork.");
            exit (2);
         }
         else if (pid != 0)
            _exit (0);
      }

      if (log_level > 0)
      {
         /*
          * Disassociate ourself from the controlling
          * terminal. Set ourself to be the new process 
          * group leader, then close fd's.
          */
         if (setsid() < 0)
            log_conn(log_level,
                  "ERROR: Could not disassociate from controlling terminal.");

         /*
          * close all file descriptors and 
          * open them pointing to /dev/null;
          */
         close(STDOUT_FILENO);
         open("/dev/null", O_WRONLY);
         dup2(STDOUT_FILENO, STDERR_FILENO);
      }
      close(STDIN_FILENO);
      open("/dev/null", O_RDONLY);
   }

   return (0);
}

int main(int argc, char **argv)
{
   int             i, dofork;
   char           *device, *optstr = "c:d:nu:", c, user[50];
   uid_t           uid;
   struct passwd  *pw;
   struct stat     filestat;
   extern char    *optarg;

   /*
    * Initializations.
    */
   device     = "any";
   dofork     = 1;
   opterr     = 0;
   log_level  = 0;
   log_debug  = 0;
   log_iponly = 1;

   /*
    * Check the command line options and arguments
    */
   strncpy(user, NOBODY, sizeof(user) - 2);
   strncpy(cfgfile, DEFAULT_CFG, sizeof(cfgfile) - 2);
   while ((c = getopt(argc, argv, optstr)) != EOF)
   {
      switch (c)
      {
         case 'c':
            strncpy(cfgfile, optarg, sizeof(cfgfile) - 2);
            break;

         case 'd':
            device = (char *)optarg;
            break;

         case 'n':
            dofork = 0;
            break;

         case 'u':
            strncpy(user, optarg, sizeof(user) - 2);
            break;

         default:
            log_exit(0,
                     "Usage: %s [-c config_file] [-d listen_device] [-n <no-fork>] [-u user]\n",
                     argv[0]);
      }
   }

   /*
    * Ensure that the device doesn't have any strange characters (pure ASCII).
    * This could fail in non-ASCII locales.
    */
   {
      unsigned char   *tdev;

      tdev = (unsigned char *)device;
      while (*tdev != '\0')
      {
         if (*tdev++ > (char)0x7F)
            log_exit(0, "ERROR: invalid listen device\n");
      }
   }

   /*
    * Create the raw socket and open the network device
    * to listen.
    */
   if ((pcap = pcap_open(device, 200, 1000)) == NULL)
      log_exit(0, "ERROR: Cannot open packet capture device.");

   /*
    * We've created the raw socket and opened the network packet
    * device - we no longer need to have root permissions. Let's
    * change to an unprivliged user.
    */
   if (user[0] == 0)
      log_exit(0, "ERROR: Cannot become (null) user!");
   if ((pw = getpwnam(user)) == NULL)
      log_exit(0, "ERROR: Cannot get userid for %.16s", user);
   uid = pw->pw_uid;
   setuid(uid);
   if (geteuid() != uid)
      log_exit(0, "ERROR: Could not successfully drop root permissions.");

   /*
    * Read the configuration file after we've dropped root
    * privileges (thanks to Stuart Kemp for pointing this out!).
    */
   if (cfgfile[0] != 0)
   {
      i = 0;
      my_dostat:
      if (lstat(cfgfile, (struct stat *)&filestat) == -1)
      {
         if (errno == EINTR || errno == EAGAIN)
         {
            if (i++ < 20)
               goto my_dostat;
            else
               log_exit(0, "ERROR: Could not stat %s after %d tries - terminating.",
                        cfgfile, i);
         }
      }
      else
      {
         if (read_cfg(cfgfile) != 0)
            log_exit(0, "ERROR: Error while reading %s - terminating.", cfgfile);
      }
   }

   /*
    * Background ourself, then become the process leader
    * if we are logging to syslog (we want to keep the
    * controlling terminal if we're logging to stderr).
    */
   setup_handlers();
   background(dofork);

   /*
    * Now read and log the connections.
    */
   log_conn(log_level, "NOTICE: Connlogd starting.");
   alarm(UDP_LIFE);
   read_packets(pcap);

   /*
    * We are done - close the packet capture device and exit.
    */
   log_conn(log_level, "NOTICE: Connlogd exiting.");
   pcap_close(pcap);
   exit (0);
}



