/*******************************************************************
  functions.c  --  Copyright 1999 - 2003
                      Alec Kosky (aleckosky@hotmail.com)

Alright, here's the deal (i.e. the AK License):

You are free to use this program for commercial as well as 
   non-commercial endeavors.
You are free to modify this program to suit your needs.
You are free to redistribute unmodified copies of this program.

Basically, you can do what you like with it. However, you are not
permitted to redistribute modified copies of this program. If you
would like to see changes made to this program, please submit   
them to me (aleckosky@hotmail.com) and I will incorporate them into
future versions if I feel that they are beneficial.

********************************************************************/
#include "connlogd.h"

/*
 * Necessary global variables.
 */
extern int                  log_level, log_iponly, log_debug;
extern struct udpconn_t    *udplist;
       int                  datalink, uc_count;

/*
 * External functions.
 */
extern int d_port_find(unsigned int);
extern int s_port_find(unsigned long int, unsigned int);


/*
 * Get the hostname for the given address structure. Then do a reverse
 * lookup on it and see if everything is kosher.
 */
char *hostlookup(struct in_addr *i)
{
   static char     buf[256];
   char           *addr, *addr2;
   struct hostent *he, *he2;
   struct in_addr  sa;
   int             found = 0;

   /*
    * Get the host address
    */
   addr = (char *)inet_ntoa(*i);

   if (log_iponly != 0)
      snprintf(buf, sizeof(buf) - 2, "%.20s", addr);
   else
   {
      he = gethostbyaddr((char *)i, sizeof(struct in_addr), AF_INET);
      if (he == NULL)
         snprintf(buf, sizeof(buf) - 2, "%.20s (no hostname available)",
                  addr);
      else
      {
         he2 = gethostbyname(he->h_name);
         if (he2 == NULL)
            snprintf(buf, sizeof(buf) - 2,
                     "%.20s (no forward DNS for %.208s)", addr, he->h_name);
         else
         {
            int     addr_offset;

            /*
             * Run through the list of addresses and see if one of them
             * is one of the ones that we were given.
             */
            for (addr_offset = 0; he2->h_addr_list[addr_offset] != '\0';
                 addr_offset++)
            {
               (void)memcpy(&sa.s_addr, he2->h_addr_list[addr_offset],
                            sizeof(sa.s_addr));
               if (i->s_addr == sa.s_addr)
               {
                  if (found == 0)
                     found = 1;
                  break;
               }
            }

            /*
             * Print out an appropriate response (including a warning if
             * the reverse does not match the forward.
             */
            if (found != 0)
               snprintf(buf, sizeof(buf) - 2, "%.20s (%.230s)", addr,
                        he->h_name);
            else
            {
               addr2 = (char *)inet_ntoa(sa);
               snprintf(buf, sizeof(buf) - 2,
                        "%.20s (WARNING: reverse DNS for %.180s is %.20s)",
                        addr, he->h_name, addr2);
            }
         }
      }
   }

   return ((char *)&buf);
}

/*
 * Find out the name of the service we are dealing with.
 */
char *servlookup(unsigned short port)
{
   static char     buf[64];
   struct servent *se;

   se = getservbyport(port, "tcp");
   if (se == NULL)
      snprintf(buf, sizeof(buf) - 2, "port %d", ntohs(port));
   else
      snprintf(buf, sizeof(buf) - 2, "%.48s", se->s_name);

   return ((char *)&buf);
}

/*
 * This is the function that actually processes the argument list
 * and does the logging.
 *
 *    loglevel  <=  0:   Log to stderr
 *                              useful for logging with Dan J. Bernstein's
 *                              cyclog or multilog programs -- see his
 *                              daemontools package:
 *
 *                          ftp://koobera.math.uic.edu/www/daemontools.html
 *
 *    loglevel   >  0:   Log to syslog.
 */
int do_log(int loglevel, const char *fmt, va_list ap)
{
   char  buf[256];

   vsnprintf(buf, sizeof(buf) - 2, fmt, ap);
   if (loglevel > 0)
      syslog(loglevel, "%s", buf);
   else
   {
      fflush(stdout);
      if (strchr(buf, '\n') == NULL)
         fprintf(stderr, "%s\n", buf);
      else
         fprintf(stderr, "%s", buf);
      fflush(stderr);
   }

   return (0);
}

/*
 * Log the information.
 */
int log_conn(int loglevel, const char *fmt, ...)
{
   va_list  ap;
   int      retval;

   va_start(ap, fmt);
   retval = do_log(loglevel, fmt, ap);
   va_end(ap);

   return (retval);
}

/*
 * Log the information, then exit with an error.
 */
void log_exit(int loglevel, const char *fmt, ...)
{
   va_list  ap;

   va_start(ap, fmt);
   do_log(loglevel, fmt, ap);
   va_end(ap);

   exit (255);
}

/*
 * Log the connections, then reset the variables to defaults.
 */
int udp_logclean(struct udpconn_t *ptr)
{
   time_t          curtime;

   curtime = time(0);

   /*
    * Log the connections (if we have any)
    */
   if (ptr->conncount > 0)
      log_conn(log_level, "%d UDP connection(s) from %.80s",
               ptr->conncount, (char *)inet_ntoa(ptr->remaddr));

   /*
    * See if this UDP connection has been around for too long.
    */
   if (curtime - UDP_LIFE - 5 <= ptr->firstconn)
   {
      /*
       * Nope - let's reset the connection count.
       */
      ptr->conncount = 0;
      ptr->firstconn = time(0);
   }
   else
   {
      /*
       * It has been around for too long.
       */
      return (1);
   }

   return (0);
}

/*
 * Check the list of UDP connections and remove the ones that have
 * been in the list for too long. Log the ones that haven't.
 */
int clean_udplist(void)
{
   struct udpconn_t *ptr1, *ptr2;

   /*
    * Process all elements.
    */
   ptr1 = udplist;
   while (ptr1 != NULL)
   {
      /*
       * Log the connections and see if this UDP
       * connection has been around for too long.
       */
      if (udp_logclean(ptr1) == 0)
         ptr1 = ptr1->next;
      else
      {
         /*
          * Yep - it's been around too long. Clear it.
          */
         ptr2 = ptr1;
         if (ptr1->prev != NULL)
            ptr1->prev->next = ptr1->next;
         if (ptr1->next != NULL)
            ptr1->next->prev = ptr1->prev;
         if (ptr1 == udplist)
            udplist = ptr1->next;
         ptr1 = ptr1->next;
         free(ptr2);
         uc_count--;
      }
   }

   return (0);
}

/*
 * Update the UDP connection list with the specified host
 * and attempted connection port (destination port).
 */
int update_udplist(struct in_addr *haddr)
{
   struct udpconn_t   *ptr1;
   int                 found = 0;

   /*
    * Make sure that we don't store too many UDP connections in
    * memory - otherwise there is the possibility of a denial
    * of service attack (granted, it'd have to be number of
    * different machines sending UDP connections to the machine
    * all within a few minutes in order for this to happen, but
    * it's better to be safe).
    */
   if (uc_count >= UDP_MAX)
      return (1);

   found = 0;
   ptr1  = udplist;
   while (ptr1 != NULL)
   {
      /*
       * Find the host address in the list.
       */
      if (haddr->s_addr != (ptr1->remaddr).s_addr)
         ptr1 = ptr1->next;
      else
      {
         /*
          * Increment the connection count.
          */
         ptr1->conncount++;
         found = 1;
         ptr1 = NULL;        /* break out of the while loop */
      }
   }
   
   /*
    * Check to see if this is a new connection attempt.
    */
   if (found == 0)
   {
      /*
       * Find the last udpconn_t structure, and then add
       * this new one to the list.
       */
      ptr1 = udplist;
      if (udplist == NULL)
      {
         if ((udplist =
              (struct udpconn_t *)malloc(sizeof(struct udpconn_t))) == NULL)
            return (-1);
         udplist->prev = NULL;
         ptr1 = udplist;
      }
      else
      {
         while (ptr1->next != NULL)
            ptr1 = ptr1->next;
         if ((ptr1->next =
              (struct udpconn_t *)malloc(sizeof(struct udpconn_t))) == NULL)
            return (-1);
         ptr1->next->prev = ptr1;
         ptr1 = ptr1->next;
      }
      ptr1->next = NULL;
      ptr1->conncount = 1;
      ptr1->firstconn = time(0);
      memcpy((struct in_addr *)&(ptr1->remaddr), haddr,
             sizeof(struct in_addr));
      uc_count++;
   }

   return (0);
}

/*
 * This is a simple function to open the packet capture device,
 * determine the datalink type, and implement a packet filter to
 * allow only TCP & UDP packets.
 */
pcap_t *pcap_open(char *dev, int snaplen, int timeout)
{
   char               *device, errbuf[PCAP_ERRBUF_SIZE];
   pcap_t             *pptr;
   uint32_t            localnet, netmask;
   struct bpf_program  bpfcode;

   /*
    * If there was no device specified, look it up.
    */
   device = dev;
   if (device == NULL)
   {
      if ((device = pcap_lookupdev(errbuf)) == NULL)
      {
         log_conn(log_level,
                  "ERROR: Could not determine packet capture device: %.80s",
                  errbuf);
         return (NULL);
      }
   }

   /*
    * Get the local network and netmask for pcap_compile
    * (in case of broadcasts & such).
    */
   if (pcap_lookupnet(device, &localnet, &netmask, errbuf) < 0)
   {
      log_conn(log_level,
               "ERROR: Could not determine local network information: %s",
               errbuf);
      return (NULL);
   }

   /*
    * Open the packet capture device.
    */
   if ((pptr = pcap_open_live(device, snaplen, 0, timeout, 
                              errbuf)) == NULL)
   {
      log_conn(log_level, "ERROR: pcap_open error: %.80s", errbuf);
      return (NULL);
   }

   /*
    * Determine what kind of datalink we're using (it affects
    * header offsets).
    */
   if ((datalink = pcap_datalink(pptr)) < 0)
      log_conn(log_level, "ERROR: Could not determine datalink type: %.80s",
               pcap_geterr(pptr));

   /*
    * Create a filter to allow only TCP & UDP packets.
    */
   if (pcap_compile(pptr, &bpfcode, "tcp or udp", 0 , netmask) < 0)
   {
      log_conn(log_level,
               "ERROR: Could not create packet capture filter: %.80s",
               pcap_geterr(pptr));
      pcap_close(pptr);
      return (NULL);
   }

   /*
    * Finally, instantiate the filter.
    */
   if (pcap_setfilter(pptr, &bpfcode) < 0)
   {
      log_conn(log_level,
               "ERROR: Could not instantiate packet capture filter: %.80s",
               pcap_geterr(pptr));
      pcap_close(pptr);
      return (NULL);
   }

   /*
    * We're outta here - nothing left to be done.
    */
   return (pptr);
}

int read_packets(pcap_t *pc)
{
   int                 i, len, hlen, offset = 0;
   char               *ptr, *addr, charbuf[24];
   struct ip          *ihdr;
   struct tcpiphdr    *thdr;
   struct udpiphdr    *uhdr;
   struct pcap_pkthdr  hdr;
   struct flag_list_t  flag_list[] =
   {
      { TH_SYN,  "SYN" },
      { TH_FIN,  "FIN" },
      { TH_URG,  "URG" },
      { TH_PUSH, "PUSH" },
      { TH_RST,  "RST" },
      { TH_B64,  "BOGUS64" },
      { TH_B128, "BOGUS128" },
      { 0, "" }
   };  
      
   /*
    * Read the headers from each packet coming in.
    */
   while (1)
   {
      /*
       * Wait for the packet to become ready
       */
      while ((ptr = (char *)pcap_next(pc, &hdr)) == NULL)
         usleep(500);

      /*
       * We don't really care about the datalink headers and such.
       * We'll just skip over these...
       */
      len = hdr.caplen;
      switch (datalink)
      {
         case DLT_NULL:      /* null header = 4 bytes */
            offset = 4;
            break;

         case DLT_EN10MB:    /* ether_header = 14 bytes */
         case DLT_IEEE802:
            offset = sizeof(struct ether_header);
            break;

         case DLT_SLIP:      /* SLIP & PPP both use a 24 byte header */
         case DLT_PPP:
            offset = 24;
            break;

         default:
            /*
             * Don't want to waste CPU cycles on 
             * protocols that we don't know about.
             */
            ptr = NULL;
      }

      /*
       * We have valid data (don't want to encounter null-pointer errors).
       */
      if (ptr != NULL)
      {
         ihdr = (struct ip *)(ptr+offset);
         if (ihdr->ip_v == IPVERSION)
         {
            /*
             * We are sure that we are dealing with the correct version
             * of IP (for now, IPv4).
             */
            hlen = ihdr->ip_hl * 4;
            if (ihdr->ip_p == IPPROTO_TCP)
            {
               /*
                * OK - we have a TCP packet. Let's verify that we
                * have complete IP & TCP headers, then proceed from
                * there.
                */
               thdr = (struct tcpiphdr *)ihdr;
               if (len >= hlen + sizeof(struct tcphdr))
               {

                  if (log_debug > 0)
                  {
                     /*
                      * Log more information about this incoming packet.
                      */
                     log_conn(log_level,
                       "TCP connection from %.20s to port %d (from port %d)",
                             inet_ntoa(ihdr->ip_src), ntohs(thdr->ti_dport),
                             ntohs(thdr->ti_sport));

                     if (log_debug > 1)
                     {
                        /*
                         * Extra verbose logging.
                         */
                        log_conn(log_level,
                                 "TCP Seq #%lu   Ack #%lu   Flags: 0x%.1x", 
                                 ntohl(thdr->ti_seq), ntohl(thdr->ti_ack),
                                 thdr->ti_flags);
                     }
                  }

                  /*
                   * Check for the various types of TCP connections
                   * and probes (first ensure that we aren't supposed to
                   * ignore this connection).
                   */
                  addr = (char *)inet_ntoa(ihdr->ip_src);
                  if (d_port_find(ntohs(thdr->ti_dport)) != 0 &&
                      s_port_find(0, ntohs(thdr->ti_sport)) != 0 &&
                      s_port_find(ihdr->ip_src.s_addr, 
                                  ntohs(thdr->ti_sport)) != 0)
                  {
                     /*
                      * Cleanup the buffer that will contain the list of
                      * flags that were set in the header.
                      */
                     memset(charbuf, 0, sizeof(charbuf));
                     if (!CONN_ACK(thdr->ti_flags))       /* TH_ACK not set */
                     {
                        if (CONN_NULL(thdr->ti_flags))
                           strncpy(charbuf, "NULL", sizeof(charbuf) - 2);
                        else 
                        {
                           for (i = 0; flag_list[i].flag != 0; i++)
                           {
                              if (thdr->ti_flags & flag_list[i].flag)
                              {
                                 if (charbuf[0] == 0)
                                    strncpy(charbuf, flag_list[i].comment,
                                            sizeof(charbuf) - 2);
                                 else
                                 {
                                    strncat(charbuf, "+",
                                            sizeof(charbuf) -
                                            strlen(charbuf) - 2);
                                    strncat(charbuf, flag_list[i].comment,
                                            sizeof(charbuf) -
                                            strlen(charbuf) - 2);
                                 }
                              }
                           }
                        }

                        /*
                         * Log the connection attempt.
                         */
                        {
                           unsigned char    wmesg[48];

                           if (thdr->ti_win < (1024 * 48))
                              wmesg[0] = (unsigned char)NULL;
                           else
                              snprintf(wmesg, sizeof(wmesg) - 2,
                                       " TCP window: %d", thdr->ti_win);

                           log_conn(log_level,
                   "%.32s connection (%.40s) attempt from %.80s port %d %.24s",
                                    servlookup(thdr->ti_dport), charbuf,
                                    hostlookup((struct in_addr *)&ihdr->ip_src),
                                    thdr->ti_sport, wmesg);
                        }
                     }
                  }
               }
            }
            else if (ihdr->ip_p == IPPROTO_UDP)
            {
               /*
                * This is a UDP packet - make sure that we
                * have complete IP & UDP headers.
                */
               if (len >= hlen + sizeof(struct udphdr))
               {
                  uhdr = (struct udpiphdr *)ihdr;
                  addr = (char *)inet_ntoa(ihdr->ip_src);

                  if (log_debug > 0)
                  {
                     log_conn(log_level,
                       "UDP connection from %.20s to port %d (from port %d)",
                              addr, ntohs(uhdr->ui_dport),
                              ntohs(uhdr->ui_sport));
                  }

                  /*
                   * We don't do name lookups on UDP packets - it
                   * can cause infinite loops if we're looking up
                   * the DNS server (I'll fix this in the future
                   * when ignore lists are implemented, but until
                   * then...)
                   */
                  if (d_port_find(uhdr->ui_dport) != 0 &&
                      s_port_find(0, uhdr->ui_sport) != 0 &&
                      s_port_find(ihdr->ip_src.s_addr,
                                  uhdr->ui_sport) != 0)
                  {
                     /*
                      * We used to log when UDP connection attempts became too
                      * numerous, but it could lead to an exhaustion of
                      * logging resources (disk, memory).
                      */
                     update_udplist((struct in_addr *)&(ihdr->ip_src));
                  }
               }
            }
            else
            {
               /*
                * I don't know what the hell this packet is!
                * However, because we are telling the pcap libs
                * to filter all packets but UDP & TCP, we should
                * never get here...
                */
               log_conn(log_level,
                        "WARNING: Unknown connection type (%d) from %.20s",
                        ihdr->ip_p, inet_ntoa(ihdr->ip_src));
            }
         }
      }
   }

   return (0);
}



